use super::messages::{ReceiveMsg, SendMsg};
use std::sync::mpsc::Receiver;
use uuid::Uuid;

/// MoreThanText database Cache
pub struct Cache {
    data: Vec<Uuid>,
    rx: Receiver<SendMsg>,
}

impl Cache {
    /// Create database cache
    ///
    /// This should not be called directly.
    /// It is part of the MoreThanText::new function.
    pub fn new(rx: Receiver<SendMsg>) -> Self {
        Self {
            data: Vec::new(),
            rx: rx,
        }
    }

    /// Starts listening for database requests.
    ///
    /// Should not be directly called.
    /// Part of the MoreThanText::new function.
    pub fn listen(&mut self) {
        loop {
            match self.rx.recv().unwrap() {
                SendMsg::OpenSession(msg) => msg.tx.send(self.get_session(msg.id)).unwrap(),
            }
        }
    }

    fn get_session(&mut self, id: Option<String>) -> ReceiveMsg {
        let sid: ReceiveMsg;
        match id {
            Some(input) => match Uuid::parse_str(&input) {
                Ok(vid) => {
                    if self.data.contains(&vid) {
                        sid = ReceiveMsg::Session(vid);
                    } else {
                        sid = self.new_session();
                    }
                }
                Err(_) => sid = self.new_session(),
            },
            None => sid = self.new_session(),
        }
        sid
    }

    fn new_session(&mut self) -> ReceiveMsg {
        let mut id = Uuid::new_v4();
        while self.data.contains(&id) {
            id = Uuid::new_v4();
        }
        self.data.push(id.clone());
        ReceiveMsg::Session(id)
    }
}

#[cfg(test)]
mod session {
    use super::*;
    use std::sync::mpsc::channel;

    #[test]
    fn unique_ids() {
        let (_, rx) = channel();
        let mut cache = Cache::new(rx);
        let mut ids: Vec<Uuid> = Vec::new();
        for _ in 1..10 {
            let id = cache.get_session(None);
            match id {
                ReceiveMsg::Session(sid) => {
                    if ids.contains(&sid) {
                        assert!(false, "{} was a duplicate id.", sid)
                    }
                    ids.push(sid);
                }
            }
        }
    }

    #[test]
    fn existing_ids_are_reused() {
        let (_, rx) = channel();
        let mut cache = Cache::new(rx);
        let id1: Uuid;
        let id2: Uuid;
        match cache.get_session(None) {
            ReceiveMsg::Session(sid) => id1 = sid,
        }
        match cache.get_session(Some(id1.to_string())) {
            ReceiveMsg::Session(sid) => id2 = sid,
        }
        assert_eq!(id2, id1);
    }

    #[test]
    fn bad_ids_generate_new_ones() {
        let (_, rx) = channel();
        let mut cache = Cache::new(rx);
        let id: Uuid;
        let bad_id = "A very bad id";
        match cache.get_session(Some(bad_id.to_string())) {
            ReceiveMsg::Session(sid) => id = sid,
        }
        assert_ne!(id.to_string(), bad_id);
    }

    #[test]
    fn expired_ids_generate_new_ids() {
        let (_, rx) = channel();
        let mut cache = Cache::new(rx);
        let old_id = Uuid::new_v4();
        let id: Uuid;
        match cache.get_session(Some(old_id.to_string())) {
            ReceiveMsg::Session(sid) => id = sid,
        }
        assert_ne!(id, old_id);
    }
}
